# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import json
import logging
import os
import ssl
import sys
from websocket import create_connection
import Queue
import traceback

log = logging.getLogger(__name__)


class WebSocket():
    """ WebSocket class implementation

    """

    def __init__(self, url):
        self.url = url
        self.connect_soc()

    def connect_soc(self):
        try:
            self.WebSocketConnection(self.url, False)
        except Exception:
            log.error(traceback.format_exc())
            raise Exception("WebSocket exception, please check the WebSocket address and try again.")

    # CERTPATH to be taken in consideration later
    def WebSocketConnection(self, host, verify_cert=True, CERTPATH="/etc/liota/cert"):
        if not verify_cert:
            self.ws = None
            self.ws = create_connection(host, enable_multithread=True,
                                        sslopt={"cert_reqs": ssl.CERT_NONE})
        else:
            self.ws = None
            if os.path.isfile(CERTPATH):
                try:
                    self.ws = create_connection(host, enable_multithread=True,
                                                sslopt={"cert_reqs": ssl.CERT_REQUIRED,
                                                        "ca_certs": CERTPATH})
                except ssl.SSLError:
                    pass
            if self.ws is None:
                raise (IOError("Couldn't verify host certificate"))

    def receive(self, queue):
        try:
            log.info("Stream Opened")
            while True:
                msg = self.ws.recv()
                log.debug("Message received while running {0}".format(msg))
                if msg is "":
                    log.error(traceback.format_exc())
                    log.error("Stream Closed")
                    raise Exception("No message received from the server, please check the connection.")
                log.debug("RX {0}".format(msg))
                queue.put(msg)
        except Exception:
            self.close()

    def send(self, msg):
        log.debug("Sending data to DCC")
        log.debug("TX Sending message {0}".format(msg))
        try:
            self.ws.send(msg)
        except:
            # TODO: Retry logic required to be re-designed
            attempts = 1
            while attempts < 4:
                try:
                    log.debug("Exception while sending data, applying retry logic.")
                    self.connect_soc()
                    log.info("Created New Websocket")
                    log.debug("TX Sending message {0}".format(msg))
                    self.ws.send(msg)
                    break
                except:
                    # Three times retry websocket connection for publishing data
                    log.info("{0} attempt".format(attempts))
                    attempts += 1
                    if attempts == 4:
                        self.close()
                        log.error(traceback.format_exc())
                        raise Exception("Exception while sending data, please check the connection and try again.")
                    else:
                        log.exception("Exception while sending data, please check the connection and try again.")
                        self.close()

    def close(self):
        if self.ws is not None:
            self.ws.close()
        log.debug("Connection closed, cleanup done")



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging
from numbers import Number

from aenum import UniqueEnum

from liota.lib.utilities.filters.filter import Filter

log = logging.getLogger(__name__)


class Type(UniqueEnum):
    """
    Enum for different filter-types supported by RangeFilter

    ** Accept Filters:
        *  CLOSED       - [1, 10] values from 1...10 are accepted
        *  OPEN         - (1, 10) values from 2...9 are accepted
        *  CLOSED_OPEN  - [1, 10) values from 1...9 are accepted
        *  OPEN_CLOSED  - (1, 10] values from 2...10 are accepted

    ** Reject Filters:
        *  CLOSED_REJECT        - [1, 10] values from 1...10 are rejected
        *  OPEN_REJECT          - (1, 10) values from 2...9 are rejected
        *  CLOSED_OPEN_REJECT   - [1, 10) values from 1...9 are rejected
        *  OPEN_CLOSED_REJECT   - (1, 10] values from 2...10 are rejected

    ** Filters bounded at one end:
        *  LESS_THAN    -  values < 1 are accepted
        *  AT_MOST      -  values <= 1 are accepted
        *  GREATER_THAN -  values > 10 are accepted
        *  AT_LEAST     -  values >= 10 are accepted
    """
    #  Accept filters - bounded at both ends
    CLOSED = 0
    OPEN = 1
    CLOSED_OPEN = 2
    OPEN_CLOSED = 3
    #  Reject filters - bounded at both ends
    CLOSED_REJECT = 4
    OPEN_REJECT = 5
    CLOSED_OPEN_REJECT = 6
    OPEN_CLOSED_REJECT = 7
    #  Filters - bounded at one end
    LESS_THAN = 8
    AT_MOST = 9
    GREATER_THAN = 10
    AT_LEAST = 11


class RangeFilter(Filter):
    """
    A simple lightweight filter, that filters values based on the specified filter type (range).
    """

    def __init__(self, filter_type, lower_bound, upper_bound):
        """
        :param filter_type: Any one type from Filter Enum.
        :param lower_bound: Lower bound
        :param upper_bound: Upper bound
        """
        if filter_type not in Type:
            log.error("Unsupported Filter Type.")
            raise TypeError("Unsupported Filter Type.")

        self.filter_type = filter_type
        self._validate(lower_bound, upper_bound)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def _validate(self, lower_bound, upper_bound):
        """
        Validation of lower_bound and upper_bound value for the specified filter_type.

        :param lower_bound: Lower bound
        :param upper_bound: Upper bound
        :return: None
        """

        if self.filter_type in [Type.CLOSED, Type.OPEN, Type.CLOSED_OPEN, Type.OPEN_CLOSED,
                                Type.CLOSED_REJECT, Type.OPEN_REJECT, Type.CLOSED_OPEN_REJECT,
                                Type.OPEN_CLOSED_REJECT] and \
                (not isinstance(lower_bound, Number) or not isinstance(upper_bound, Number)):
            log.error("Both lower_bound and upper_bound must be a number")
            raise TypeError("Both lower_bound and upper_bound must be a number")

        elif self.filter_type in [Type.LESS_THAN, Type.AT_MOST] and not isinstance(lower_bound, Number):
            log.error("lower_bound must be a number")
            raise TypeError("lower_bound must be a number")

        elif self.filter_type in [Type.GREATER_THAN, Type.AT_LEAST] and not isinstance(upper_bound, Number):
            log.error("upper_bound must be a number")
            raise TypeError("upper_bound must be a number")

    def filter(self, v):
        """
        RangeFilter Implementation. Performs filtering based on specified filter_type.

        :param v: Collected value
        :return: Filtered value or None
        """
        result = None
        log.info("Applying RangeFilter " + str(self.filter_type))
        if not isinstance(v, Number):
            log.warn("Value is not a number. Returning without applying filter")
            return v

        # Accept filters - bounded at both ends
        elif self.filter_type is Type.CLOSED and (self.lower_bound <= v <= self.upper_bound):
            result = v

        elif self.filter_type is Type.OPEN and (self.lower_bound < v < self.upper_bound):
            result = v

        elif self.filter_type is Type.CLOSED_OPEN and (self.lower_bound <= v < self.upper_bound):
            result = v

        elif self.filter_type is Type.OPEN_CLOSED and (self.lower_bound < v <= self.upper_bound):
            result = v

        # Reject filters - bounded at both ends
        elif self.filter_type is Type.CLOSED_REJECT and (not self.lower_bound <= v <= self.upper_bound):
            result = v

        elif self.filter_type is Type.OPEN_REJECT and (not self.lower_bound < v < self.upper_bound):
            result = v

        elif self.filter_type is Type.CLOSED_OPEN_REJECT and (not self.lower_bound <= v < self.upper_bound):
            result = v

        elif self.filter_type is Type.OPEN_CLOSED_REJECT and (not self.lower_bound < v <= self.upper_bound):
            result = v

        # Filters - bounded at one end
        elif self.filter_type is Type.LESS_THAN and (v < self.lower_bound):
            result = v

        elif self.filter_type is Type.AT_MOST and (v <= self.lower_bound):
            result = v

        elif self.filter_type is Type.GREATER_THAN and (v > self.upper_bound):
            result = v

        elif self.filter_type is Type.AT_LEAST and (v >= self.upper_bound):
            result = v

        if result is not None:
            log.info("Value passed by filter : %s" % str(result))
        else:
            log.info("Value rejected by filter : %s" % str(v))

        return result



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging

log = logging.getLogger(__name__)


class Identity:
    """
    This class encapsulates certificates and credentials related to a connection both at Dcc and Device side.
    """

    def __init__(self, root_ca_cert, username, password, cert_file, key_file):
        """
        :param root_ca_cert: Root CA certificate path or Self-signed server certificate path
        :param username: Username
        :param password: Corresponding password
        :param cert_file: Device certificate file path
        :param key_file: Device certificate key-file path
        """
        self.root_ca_cert = root_ca_cert
        self.username = username
        self.password = password
        self.cert_file = cert_file
        self.key_file = key_file
        log.debug("Created Identity with rootCA path: {0}, username: {1}, device_cert_path: {2}"
                  "device_key_file_path: {3}".format(self.root_ca_cert, self.username, self.cert_file,
                                                     self.key_file))



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#
import json
import logging
import time
from threading import Thread

from coapthon.server.coap import CoAP
from coapthon.resources.resource import Resource
from liota.disc_listeners.discovery_listener import DiscoveryListener
from liota.lib.utilities.utility import LiotaConfigPath

logg = logging.getLogger(__name__)
LiotaConfigPath().setup_logging()

class MsgResource (Resource):

    def __init__(self, disc=None, name="MsgResource", coap_server=None):
        super(MsgResource, self).__init__(name, coap_server, visible=True,
                                            observable=True, allow_children=True)
        self.payload = "Msg Resource"
        self.resource_type = "rt1"
        self.content_type = "text/plain"
        self.interface_type = "if1"
        self.disc = disc

    def render_PUT(self, request):
        logg.debug('PUT payload: {0}'.format(request.payload))
        # process received device messages'
        # better create another thread to process received messages
        try:
            payload = json.loads(request.payload)

            Thread(target=self.proc_dev_msg, name="CoapMsgProc_Thread", args=(payload, self.disc, )).start()
        except ValueError, err:
            # json can't be parsed
            logg.error('Value: {0}, Error:{1}'.format(request.payload, str(err)))

        self.edit_resource(request)
        return self

    def proc_dev_msg(self, payload, disc):
        logg.debug("proc_dev_msg payload:{0}".format(payload))
        disc.device_msg_process(payload)

    def render_GET(self, request):
        return self

class CoAPServer(CoAP):
    def __init__(self, host, port, multicast=False, disc=None):
        CoAP.__init__(self, (host, port), multicast)
        self.disc = disc
        self.add_resource('message/', MsgResource(disc=disc))

class CoapListener(DiscoveryListener):
    """
    CoapListener does inter-process communication (IPC), and
    waits to receive messages sent by devices.
    """

    def __init__(self, ip_port, name=None, discovery=None):
        super(CoapListener, self).__init__(name=name)
        str_list = ip_port.split(':')
        if str_list[0] == "" or str_list[0] == None:
            logg.warning("No ip is specified!")
            self.ip = "127.0.0.1"
        else:
            self.ip = str(str_list[0])
        if str_list[1] == "" or str_list[1] == None:
            logg.error("No port is specified!")
            self.port = 5683
        else:
            self.port = int(str_list[1])
        self.discovery = discovery
        self.flag_alive = True
        logg.debug("CoapListener is initialized")
        print "CoapListener is initialized"
        self.start()

    def run(self):
        if self.flag_alive:
            logg.info('CoapListerner is running')
            print 'CoapListerner is running'

            # start coap server
            self.server = CoAPServer(self.ip, self.port, multicast=False, disc=self.discovery)
            try:
                self.server.listen(10)
                while self.flag_alive:
                    time.sleep(100)
                logg.info("Thread exits: %s" % str(self.name))
            except KeyboardInterrupt:
                print "Server Shutdown"
                self.server.close()
                print "Exiting..."
        else:
            logg.info("Thread exits: %s" % str(self.name))
            print "Thread exits:", str(self.name)

    def clean_up(self):
        self.flag_alive = False
        if self.server is not None:
            self.server.close()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from liota.core.package_manager import LiotaPackage

dependencies = ["iotcc", "examples/bike_simulator"]


def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate


class PackageClass(LiotaPackage):

    def create_udm(self, bike_model):
        ureg = bike_model.ureg

        import time
        import math

        #-------------------------------------------------------------------
        # The following functions operate on physical variables represented
        # in pint objects, and returns a pint object, too.
        # Decorators provided by the pint library are used to check the
        # dimensions of arguments passed to the functions.

        @ureg.check(ureg.rpm, ureg.m)
        def get_speed(revolution, radius):
            return revolution * radius

        @ureg.check(ureg.m / ureg.sec)
        @static_vars(speed_last=None, time_last=None)
        def get_acceleration(speed):
            t = time.time()
            if get_acceleration.time_last is None:
                acc = 0 * ureg.m / ureg.sec ** 2
            else:
                acc = (speed - get_acceleration.speed_last) / \
                      ((t - get_acceleration.time_last) * ureg.sec)
            get_acceleration.speed_last = speed
            get_acceleration.time_last = t
            return acc

        @ureg.check(ureg.m ** 2, ureg.m / ureg.sec, ureg.kg / ureg.m ** 3)
        def get_resistance(area, speed, k):
            return (k * area * speed ** 2)

        @ureg.check(ureg.kg, ureg.m / ureg.sec ** 2)
        def get_force(mass, acceleration):
            return mass * acceleration

        @ureg.check(ureg.newton, ureg.m / ureg.sec)
        def get_power(force, speed):
            return force * speed

        #-------------------------------------------------------------------
        # This is a sampling method, which queries the physical model, and
        # calls the physical functions to calculate a desired variable.

        def get_bike_speed():
            speed = get_speed(
                bike_model.get_revolution(),
                bike_model.get_radius_wheel()
            ).to(ureg.m / ureg.sec)
            return speed.magnitude

        #-------------------------------------------------------------------
        # This is a more complex sampling method, which queries the physical
        # model.

        def get_bike_power():
            weight_total = bike_model.get_weight_bike() + \
                bike_model.get_weight_rider() + \
                bike_model.get_weight_load()
            speed = get_speed(
                bike_model.get_revolution(),
                bike_model.get_radius_wheel()
            )
            power_acceleration = get_power(
                get_force(
                    weight_total,
                    get_acceleration(speed)
                ),
                speed
            ).to(ureg.watt)
            power_gravity = get_power(
                get_force(
                    weight_total,
                    9.8 * ureg.m / ureg.sec ** 2
                ),
                speed * math.sin(bike_model.get_slope())
            ).to(ureg.watt)
            power_resistance = get_power(
                get_resistance(
                    bike_model.get_area(),
                    speed,
                    10 * ureg.kg / ureg.m ** 3
                ).to(ureg.newton),
                speed
            ).to(ureg.watt)
            power = power_acceleration + power_gravity + power_resistance
            return power.to(ureg.watt).magnitude

        self.get_bike_speed = get_bike_speed
        self.get_bike_power = get_bike_power

    def run(self, registry):
        from liota.entities.metrics.metric import Metric
        import copy

        # Acquire resources from registry
        self.config_path = registry.get("package_conf")
        self.iotcc = registry.get("iotcc")
        self.iotcc_edge_system = copy.copy(registry.get("iotcc_edge_system"))
        bike_simulator = registry.get("bike_simulator")

        self.iotcc_bike = self.iotcc.register(bike_simulator)
        self.iotcc.create_relationship(self.iotcc_edge_system, self.iotcc_bike)

        ureg = bike_simulator.ureg
        self.create_udm(bike_model=bike_simulator)

        # Create metrics
        self.metrics = []

        metric_name = "model.bike.speed"
        bike_speed = Metric(
            name=metric_name,
            unit=(ureg.m / ureg.sec),
            interval=5,
            sampling_function=self.get_bike_speed
        )
        reg_bike_speed = self.iotcc.register(bike_speed)
        self.iotcc.create_relationship(self.iotcc_bike, reg_bike_speed)
        reg_bike_speed.start_collecting()
        self.metrics.append(reg_bike_speed)

        metric_name = "model.bike.power"
        bike_power = Metric(
            name=metric_name,
            unit=ureg.watt,
            interval=5,
            sampling_function=self.get_bike_power
        )
        reg_bike_power = self.iotcc.register(bike_power)
        self.iotcc.create_relationship(self.iotcc_bike, reg_bike_power)
        reg_bike_power.start_collecting()
        self.metrics.append(reg_bike_power)

        # Use the iotcc_device_name as identifier in the registry to easily refer the registered device in other packages
        registry.register("iotcc_bike_simulated", self.iotcc_bike)


    def clean_up(self):

        # Get values from configuration file
        config = {}
        execfile(self.config_path + '/sampleProp.conf', config)

        for metric in self.metrics:
            metric.stop_collecting()

        #Unregister iotcc device
        if config['ShouldUnregisterOnUnload'] == "True":
            self.iotcc.unregister(self.iotcc_bike)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging
import inspect
import imp
import os
import fcntl
import stat
import re
import hashlib
import ConfigParser
from threading import Thread, Lock
from Queue import Queue
from time import sleep
from abc import ABCMeta, abstractmethod

from liota.lib.utilities.utility import LiotaConfigPath, read_liota_config

log = logging.getLogger(__name__)

if __name__ == "__main__":
    log.warning("Package manager is not supposed to run alone")

    import liota.core.package_manager as actual_package_manager

    log.debug("MainThread is waiting for interruption signal...")
    try:
        while not isinstance(
            actual_package_manager.package_thread,
            actual_package_manager.PackageThread
        ) or actual_package_manager.package_thread.isAlive():
            sleep(1)
    except (KeyboardInterrupt, SystemExit):
        pass
    finally:
        if actual_package_manager.package_message_queue is not None:
            actual_package_manager.package_message_queue.put(["terminate"])
    log.info("Exiting Mainthread")
    exit()

is_package_manager_initialized = False

package_message_queue = None
package_messenger_thread = None
package_thread = None
package_lock = None
package_path = None
package_messenger_pipe = None

# Parse Liota configuration file
package_path = os.path.abspath(
    read_liota_config('PKG_CFG', 'pkg_path')
)
package_messenger_pipe = os.path.abspath(
    read_liota_config('PKG_CFG', 'pkg_msg_pipe')
)
assert(isinstance(package_path, basestring))
assert(isinstance(package_messenger_pipe, basestring))

package_startup_list_path = None
package_startup_list = []

# Parse packages to load at start-up
package_startup_list_path = os.path.abspath(
    read_liota_config('PKG_CFG', 'pkg_list')
)

class ResourceRegistryPerPackage:
    """
    ResourceRegistryPerPackage creates temporary objects for packages while
    loading, so when they register their resource refs, we can keep track of
    them in resource registry and automatically remove them later when these
    packages are unloaded.
    """

    def __init__(self, outer, package_name):
        self._outer = outer
        self._package_name = package_name

    def register(self, identifier, ref):
        self._outer.register(identifier, ref, self._package_name)

    def get(self, identifier):
        return self._outer.get(identifier)

    def has(self, identifier):
        return self._outer.has(identifier)


class ResourceRegistry:
    """
    ResourceRegistry is a wrapped structure for Liota packages to register
    resources and find resources registered by other packages.
    """

    def __init__(self):
        self._registry = {}  # key: resource name, value: resource ref
        self._packages = {}  # key: package name, value: list of resource names

    def register(self, identifier, ref, package_name=None):
        if identifier in self._registry:
            raise KeyError("Conflicting resource identifier: " + identifier)
        self._registry[identifier] = ref
        if package_name:
            if package_name not in self._packages:
                self._packages[package_name] = []
            self._packages[package_name].append(identifier)

    def deregister(self, identifier):
        del self._registry[identifier]

    def get(self, identifier):
        return self._registry[identifier]

    def has(self, identifier):
        return identifier in self._registry

    #-----------------------------------------------------------------------
    # This method generate a package specific registry object, so when they
    # register their resource refs, we keep track of them and can deregister
    # them automatically if package is unloaded.

    def get_package_registry(self, package_name):
        return ResourceRegistryPerPackage(self, package_name)


class LiotaPackage:
    """
    LiotaPackage is ABC (abstract base class) of all package classes.
    Developers should extend LiotaPackage class and implement the abstract methods.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def run(self, registry):
        raise NotImplementedError

    @abstractmethod
    def clean_up(self):
        raise NotImplementedError

#---------------------------------------------------------------------------
# This method calculates SHA-1 checksum of file.
# May raise IOError upon "open"


def sha1sum(path_file):
    sha1 = hashlib.sha1()
    with open(path_file, "rb") as fp:
        while True:
            data = fp.read(65536)  # buffer size
            if not data:
                break
            sha1.update(data)
    return sha1


class PackageRecord:
    """
    PackageRecord is instantiated for each package loaded.
    It contains necessary information of packages, including but may not be
    limited to file names, checksums, and refs to package class instances.
    These records make it possible to start and stop package class instances
    elegantly.
    """

    def __init__(self, file_name):
        self._file_name = file_name
        self._ext = None
        self._sha1 = None
        self._dependents = {}  # key: dependent name, value: None
        self._dependencies = []

        #-------------------------------------------------------------------
        # To guarantee successful garbage collection when record is removed,
        # this should be the ONLY variable to keep its reference.

        self._instance = None

    def set_instance(self, obj):
        if self._instance is not None:
            log.warning("Should not override instance of package class")
            return False
        self._instance = obj
        return True

    def get_instance(self):
        return self._instance

    def set_sha1(self, sha1):
        self._sha1 = sha1

    def get_sha1(self):
        return self._sha1

    def set_ext(self, ext):
        self._ext = ext

    def get_ext(self):
        return self._ext

    def get_dependents(self):
        return self._dependents.keys()

    def add_dependent(self, file_name):
        self._dependents[file_name] = None

    def del_dependent(self, file_name):
        del self._dependents[file_name]

    def get_dependencies(self):
        return self._dependencies

    def set_dependencies(self, list_dependencies):
        self._dependencies = list_dependencies


class PackageThread(Thread):
    """
    PackageThread should be instantiated only once.
    Its instance serves core functionalities of Liota package manager,
    including load/unload of packages, maintenance of loaded package records,
    etc.
    """

    def __init__(self, name=None):
        Thread.__init__(self, name=name)

        global package_path

        self._packages_loaded = {}  # key: package name, value: PackageRecord obj
        self._resource_registry = ResourceRegistry()
        self._resource_registry.register("package_conf", package_path)
        self.flag_alive = True
        self.start()

    #-----------------------------------------------------------------------
    # This method is used to handle listing commands

    def _cmd_handler_list(self, parameter):
        if parameter == "packages" or parameter == "pkg":
            log.warning("List of packages - \n\t%s"
                        % "\n\t".join(sorted(
                            self._packages_loaded.keys()
                        ))
                        )
            return
        if parameter == "resources" or parameter == "res":
            log.warning("List of resources - \n\t%s"
                        % "\n\t".join(sorted(
                            self._resource_registry._registry.keys()
                        ))
                        )
            return
        if parameter == "threads" or parameter == "th":
            import threading

            log.warning("Active threads - \n\t%s"
                        % "\n\t".join(map(
                            lambda tref: "%s: %016x %s %s" % (
                                tref.name,
                                tref.ident,
                                type(tref).__name__.split(".")[-1],
                                tref.isAlive()
                            ),
                            sorted(
                                threading.enumerate(),
                                key=lambda tref: tref.ident
                            )
                        ))
                        )
            return
        log.warning("Unsupported list")

    #-----------------------------------------------------------------------
    # This method is used to handle statistical commands

    def _cmd_handler_stat(self, parameter):
        if parameter == "metrics" or parameter == "met":
            from liota.core.metric_handler \
                import event_ds, collect_queue, send_queue, \
                CollectionThreadPool, collect_thread_pool

            stats = ["n/a", "n/a", "n/a", "n/a"]
            if isinstance(event_ds, Queue):
                stats[0] = str(event_ds.qsize())
            if isinstance(send_queue, Queue):
                stats[1] = str(send_queue.qsize())
            if isinstance(collect_queue, Queue):
                stats[2] = str(collect_queue.qsize())
            if isinstance(collect_thread_pool, CollectionThreadPool):
                stats[3] = collect_thread_pool.get_stats_working()[0]
            log.warning(("Number of metrics in - \n\t"
                         + "Waiting queue: %s\n\t"
                         + "Sending queue: %s\n\t"
                         + "Collecting queue: %s\n\t"
                         + "Collecting threads: %s"
                         ) % tuple(stats))
            return
        if parameter == "collection_threads" or parameter == "col":
            from liota.core.metric_handler \
                import CollectionThreadPool, collect_thread_pool

            stats = ["n/a", "n/a", "n/a", "n/a"]
            if isinstance(collect_thread_pool, CollectionThreadPool):
                stats = map(
                    lambda n: str(n),
                    collect_thread_pool.get_stats_working()
                )
            log.warning(("Status of collection threads - \n\t"
                         + "Collecting: %s\n\t"
                         + "Alive: %s\n\t"
                         + "Pool: %s\n\t"
                         + "Capacity: %s"
                         ) % tuple(stats))
            return
        if parameter == "threads" or parameter == "th":
            import threading

            log.warning("Count of active threads: %d"
                        % threading.active_count())
            return
        log.warning("Unsupported stat")

    #-----------------------------------------------------------------------
    # This method will loop on message queue and select methods to call with
    # respect to commands received.

    def run(self):
        global package_lock
        global package_startup_list

        # Load packages specified for automatic loading
        with package_lock:
            self._package_load_auto()

        # Listen on message queue for commands
        # Other packages are loaded here according to commands received
        while self.flag_alive:
            msg = package_message_queue.get()
            log.info("Got message in package messenger queue: %s"
                     % " ".join(msg))
            if not isinstance(msg, tuple) and not isinstance(msg, list):
                raise TypeError(type(msg))

            # Switch on message content (command), determine what to do
            command = msg[0]
            if command in ["load", "unload", "delete", "reload", "update"]:
                #-----------------------------------------------------------
                # Use these commands to handle package management tasks

                with package_lock:
                    if len(msg) < 2:
                        log.warning("No package is specified: %s" % command)
                        continue
                    if len(msg) > 2:
                        list_packages = msg[1:]
                        if command == "load":
                            self._package_load_list(list_packages)
                        elif command == "unload":
                            self._package_unload_list(list_packages)
                        elif command == "update":
                            self._package_update_list(list_packages)
                        else:
                            log.warning("Batch operation not supported: %s"
                                        % command)
                        continue
                    file_name = msg[1]
                    if command == "load":
                        self._package_load(file_name)
                    elif command == "unload":
                        self._package_unload(file_name)
                    elif command == "delete":
                        self._package_delete(file_name)
                    elif command == "reload":
                        self._package_reload(file_name)
                    elif command == "update":
                        self._package_update(file_name)
                    else:  # should not happen
                        raise RuntimeError("Command category error")
            elif command == "list":
                with package_lock:
                    if len(msg) != 2:
                        log.warning("Invalid format of command: %s" % command)
                        continue
                    self._cmd_handler_list(msg[1])
            elif command == "stat":
                with package_lock:
                    if len(msg) != 2:
                        log.warning("Invalid format of command: %s" % command)
                        continue
                    self._cmd_handler_stat(msg[1])
            elif command == "load_auto":
                with package_lock:
                    self._package_load_auto()
            elif command == "unload_all":
                with package_lock:
                    self._package_unload_list(self._packages_loaded.keys())
            elif command == "update_all":
                with package_lock:
                    self._package_update_list(self._packages_loaded.keys())
            elif command == "terminate":
                if self._terminate_all():
                    self.flag_alive = False
                    break
            else:
                log.warning("Unsupported command is dropped")
        log.info("Thread exits: %s" % str(self.name))

    #-----------------------------------------------------------------------
    # This method is called to check if specified package exists

    def _package_chk_exists(self, file_name, ext_forced=None):
        global package_path

        c_slash = "/"
        if package_path.endswith(c_slash):
            c_slash = ""
        path_file = os.path.abspath(package_path + c_slash + file_name)

        file_ext = None
        extensions = ["py", "pyc", "pyo"]
        prompt_ext_all = "py[co]?"
        if not ext_forced:
            for file_ext_ind in extensions:
                if os.path.isfile(path_file + "." + file_ext_ind):
                    file_ext = file_ext_ind
                    break
        else:
            if os.path.isfile(path_file + "." + ext_forced):
                file_ext = ext_forced
        if not file_ext:
            if not ext_forced:
                log.error("Package file not found: %s"
                          % (path_file + "." + prompt_ext_all))
            else:
                log.error("Package file not found: %s"
                          % (path_file + "." + ext_forced))
            return None, None
        path_file_ext = path_file + "." + file_ext
        log.debug("Package file found: %s" % path_file_ext)
        return path_file_ext, file_ext

    #-------------------------------------------------------------------
    # Attempt to load package module from file.
    # Supported file types are source files (.py) with highest priority,
    #                          compiled files (.pyc),
    #                      and optimized compiled files (.pyo).
    # Having .py files to have highest priority guarantees that coming
    # packages in .py format can override compiled files of its previous
    # version.

    def _package_module_load(self, file_name, path_file_ext, file_ext):

        module_loaded = None
        module_name = re.sub(r"\.", "_", file_name)
        try:
            if file_ext in ["py"]:
                module_loaded = imp.load_source(
                    module_name,
                    path_file_ext
                )
            elif file_ext in ["pyc", "pyo"]:
                module_loaded = imp.load_compiled(
                    module_name,
                    path_file_ext
                )
            else:  # should not happen
                raise RuntimeError("File extension category error")
        except Exception as err:
            log.error("Error loading module: %s" % str(err))
            return None, None

        log.debug("Loaded module: %s" % module_loaded.__name__)
        return module_loaded, module_name

    #-----------------------------------------------------------------------
    # This method is called to load package into current Liota process using
    # file_name (no_ext) as package identifier.

    def _package_load(self, file_name, ext_forced=None, check_stack=None):

        log.debug("Attempting to load package: %s" % file_name)

        # Check if specified package is already loaded
        if file_name in self._packages_loaded:
            log.warning("Package already loaded: %s" % file_name)
            return None

        path_file_ext, file_ext = self._package_chk_exists(
            file_name, ext_forced)
        if path_file_ext is None:
            return None

        # Read file and calculate SHA-1
        try:
            sha1 = sha1sum(path_file_ext)
        except IOError:
            log.error("Could not open file: %s" % path_file_ext)
            return None
        log.info("Loaded package file: %s (%s)"
                 % (path_file_ext, sha1.hexdigest()))

        #-------------------------------------------------------------------
        # Following sections do these:
        #   1)     from file path  load module,
        #   2)     from module     load class,
        #   3)     with class      create instance (object),
        #   4) and call method run of created instance.

        module_loaded, module_name = self._package_module_load(
            file_name, path_file_ext, file_ext)
        if module_loaded is None:
            return None

        #-------------------------------------------------------------------
        # Acquire dependency list and recursively load them.
        # If any dependency fails to load, current package will not load.

        dependencies = []
        if hasattr(module_loaded, "dependencies"):
            dependencies = getattr(module_loaded, "dependencies")
            if not isinstance(dependencies, list):
                log.error("Mal-formatted list of dependencies in module %s"
                          % module_loaded.__name__)
                return None

            if len(dependencies) > 0:
                log.info("Package %s depends on: %s"
                         % (file_name, " ".join(dependencies)))
                if not isinstance(check_stack, list):
                    check_stack = []
                check_stack.append(file_name)
                for dependency in dependencies:
                    if dependency in check_stack:
                        log.error("%s is not loaded, because %s depends on it"
                                  % (file_name, dependency))
                        check_stack.pop()
                        return None
                    if dependency not in self._packages_loaded:
                        self._package_load(dependency, check_stack=check_stack)
                    if dependency not in self._packages_loaded:
                        log.error("%s is not loaded, because %s failed to load"
                                  % (file_name, dependency))
                        check_stack.pop()
                        return None

                    # Add dependent record
                    dep_record = self._packages_loaded[dependency]
                    assert(isinstance(dep_record, PackageRecord))
                    dep_record.add_dependent(file_name)
                check_stack.pop()
                log.debug("Dependency check of package %s is complete"
                          % file_name)

        # Get package class from module and instantiate it
        if not hasattr(module_loaded, "PackageClass"):
            log.error("Invalid package: %s" % module_name)
            return None

        klass = getattr(module_loaded, "PackageClass")
        package_record = PackageRecord(file_name)
        if not package_record.set_instance(klass()):
            log.error("Unexpected failure initializing package")
            return None
        try:  # Run created instance
            package_record.get_instance().run(
                self._resource_registry.get_package_registry(file_name)
            )
        except Exception as er:
            log.error("Exception in initialization: %s" % str(er))
            return None
        package_record.set_sha1(sha1)
        package_record.set_ext(file_ext)
        package_record.set_dependencies(dependencies)
        self._packages_loaded[file_name] = package_record

        log.info("Package class from module %s is initialized"
                 % module_loaded.__name__)
        return package_record

    #-----------------------------------------------------------------------
    # This method is called to unload package using its file_name (no ext).
    # Use track_list to keep track of full file name (with ext) when unloading,
    # so reload can always load exactly that same file
    def _package_unload(self, file_name, track_list=None):
        log.debug("Attempting to unload package: %s" % file_name)

        # Check if specified package is already loaded
        if file_name not in self._packages_loaded:
            log.warning("Could not unload package - not loaded: %s"
                        % file_name)
            return False

        package_record = self._packages_loaded[file_name]
        assert(isinstance(package_record, PackageRecord))

        # Stop all dependents, before making any change to current package
        dependents = package_record.get_dependents()

        if len(dependents) > 0:
            log.info("Package %s is depended by: %s"
                     % (file_name, " ".join(dependents)))
            for dependent in dependents:
                if dependent in self._packages_loaded:
                    self._package_unload(dependent, track_list=track_list)
                if dependent in self._packages_loaded:
                    log.error("%s is still alive, because %s failed to unload"
                              % (file_name, dependent))
                    return False
            log.debug("Dependency check of package %s is complete"
                      % file_name)

        package_obj = package_record.get_instance()
        if not isinstance(package_obj, LiotaPackage):
            raise TypeError(type(package_obj))

        # Deregister resources
        # Unload should proceed no matter deregistration succeeds or not
        if file_name in self._resource_registry._packages:
            for identifier in self._resource_registry._packages[file_name]:
                self._resource_registry.deregister(identifier)
            del self._resource_registry._packages[file_name]
            log.debug("Deregistered resource refs for package: %s"
                      % file_name)
        else:
            log.warning("Could not deregister resource refs for package: %s"
                        % file_name)

        # Clean-up
        try:
            package_obj.clean_up()
        except Exception as er:
            log.error("Exception in clean-up: %s" % er)

        # Remove dependent item from dependencies
        log.debug("Package %s depends on: %s"
                  % (file_name, " ".join(package_record.get_dependencies())))
        for dependency in package_record.get_dependencies():
            self._packages_loaded[dependency].del_dependent(file_name)
            log.debug("Package %s is no longer a dependent of %s"
                      % (file_name, dependency))

        if isinstance(track_list, list):
            track_list.append((file_name, package_record.get_ext()))
        del self._packages_loaded[file_name]

        log.info("Unloaded package: %s" % file_name)
        return True

    def _package_delete(self, file_name):
        log.debug("Attempting to delete package: %s" % file_name)
        pass  # TODO
        log.info("Deleted package: %s" % file_name)

    #-----------------------------------------------------------------------
    # This method is called to reload package.
    # We keep track of full file name (with ext) when unloading, so reload
    # will always load exactly that same file, even if a different higher-
    # priority source file is added before reload.

    def _package_reload(self, file_name):
        log.debug("Attempting to reload package: %s" % file_name)

        # Check if specified package is already loaded
        if file_name not in self._packages_loaded:
            log.warning("Could not reload package - not loaded: %s"
                        % file_name)
            return False

        # Logic of reload
        track_list = []
        if self._package_unload(file_name, track_list=track_list):
            package_record = None
            track_list.reverse()
            log.info("Packages will be reloaded: %s"
                     % " ".join(
                         map(lambda item: item[0], track_list)
                     )
                     )
            for track_item in track_list:
                if track_item[0] in self._packages_loaded:
                    continue
                temp_record = \
                    self._package_load(track_item[0], ext_forced=track_item[1])
                if temp_record is not None:
                    if track_item[0] == file_name:
                        package_record = temp_record
                    log.info("Reloaded package: %s" % file_name)
                else:
                    log.error("Unloaded but could not reload package: %s"
                              % file_name)
            if not package_record is None:
                return package_record
            else:
                return False
        else:
            log.warning("Could not unload package: %s" % file_name)
        return False

    #-----------------------------------------------------------------------
    # This method is called to update package.
    # We keep track of full file name (without ext) when unloading.
    # The difference between this method and _package_reload is:
    #   1)  If target package is not loaded, this method tries to load it.
    #   2)  For all packages involved in update, this method calls
    #       _package_load to look for source files and compiled files in our
    #       preferred priority order, so updated source file can be used to
    #       update target package even if it was loaded using compiled file.

    def _package_update(self, file_name):
        log.debug("Attempting to update package: %s" % file_name)

        # Check if specified package is already loaded
        if file_name not in self._packages_loaded:
            log.info("Package is not loaded, will try to load: %s"
                     % file_name)
            return self._package_load(file_name)

        # Logic of reload
        track_list = []
        if self._package_unload(file_name, track_list=track_list):
            package_record = None
            track_list.reverse()
            log.info("Packages will be reloaded and updated: %s"
                     % " ".join(
                         map(lambda item: item[0], track_list)
                     )
                     )
            for track_item in track_list:
                if track_item[0] in self._packages_loaded:
                    continue
                temp_record = \
                    self._package_load(track_item[0])
                if temp_record is not None:
                    if track_item[0] == file_name:
                        package_record = temp_record
                    log.info("Reloaded and updated package: %s" % file_name)
                else:
                    log.error("Unloaded but could not reload package: %s"
                              % file_name)
            if not package_record is None:
                return package_record
            else:
                return False
        else:
            log.warning("Could not unload package: %s" % file_name)
        return False

        log.info("Reloaded and updated package: %s" % file_name)

    #-----------------------------------------------------------------------
    # This method is called to load a list of packages.
    # If any package in list is already loaded, it will be simply ignored -
    # Neither will this method throw an exception, nor will this package be
    # reloaded.
    # It returns True if all packages in list are successfully loaded.

    def _package_load_list(self, package_list):
        log.debug("Attempting to load packages: %s"
                  % " ".join(package_list))
        list_failed = []
        for file_name in package_list:
            if file_name in self._packages_loaded:
                continue
            if not self._package_load(file_name):
                list_failed.append(file_name)

        if len(list_failed) > 0:
            log.warning("Some packages specified in list failed to load: %s"
                        % " ".join(list_failed))
        else:
            log.info("Batch load successful")
        return len(list_failed) < 1

    #-----------------------------------------------------------------------
    # This method is called to unload a list of packages.
    # If any package in list is not loaded, it will be simply ignored -
    # This method will simply assume it is successfully unloaded and won't
    # throw an exception.
    # It returns True if all packages in list are successfully unloaded.

    def _package_unload_list(self, package_list, track_list=None):
        log.debug("Attempting to unload packages: %s"
                  % " ".join(package_list))
        list_failed = []
        for file_name in package_list:
            if file_name not in self._packages_loaded:
                continue
            if not self._package_unload(file_name, track_list=track_list):
                list_failed.append(file_name)

        if len(list_failed) > 0:
            log.warning("Some packages specified in list failed to unload: %s"
                        % " ".join(list_failed))
        else:
            log.info("Batch unload successful")
        return len(list_failed) < 1

    #-----------------------------------------------------------------------
    # This method is called to update a list of packages.
    # It first unload packages in that list, and then load them back.

    def _package_update_list(self, package_list):
        log.debug("Attempting to update packages: %s"
                  % " ".join(package_list))
        flag_failed = False

        # Acquire a list of all dependents of these packages
        track_list = []
        if not self._package_unload_list(package_list, track_list=track_list):
            flag_failed = True
        track_list.reverse()

        # Load packages, in case some packages not loaded are to be updated
        if not self._package_load_list(package_list):
            flag_failed = True

        # Load all dependents
        if len(track_list) > 0:
            if not self._package_load_list(map(lambda x: x[0], track_list)):
                flag_failed = True

        if flag_failed:
            log.warning("Some packages failed to update. See log for details")
        else:
            log.info("Batch update successful")
        return not flag_failed

    #-----------------------------------------------------------------------
    # This method is called to load packages that are specified in
    # configuration for automatic loading at start-up.

    def _package_load_auto(self):
        global package_startup_list_path
        global package_startup_list

        package_startup_list = []

        # Validate start-up list
        if isinstance(package_startup_list_path, basestring):
            try:
                with open(package_startup_list_path, "r") as fp:
                    package_startup_list = fp.read().split()
            except IOError:
                log.warning("Could not load start-up list from: %s"
                            % package_startup_list_path)
        else:
            log.info("No package is automatically loaded at start-up")

        # Load packages in a batch
        if len(package_startup_list) > 0:
            self._package_load_list(package_startup_list)

    #-----------------------------------------------------------------------
    # This method is called to delete package. All source files and compiled
    # files associated with specified package name will be affected.
    # It tries to move "deleted" files into a subdirectory called "stash".
    # If moving fails, it then tries to really delete these files.

    def _package_delete(self, file_name):
        global package_path

        log.debug("Attempting to delete package: %s" % file_name)

        # Check if specified package is loaded, if yes, unload first
        if file_name in self._packages_loaded:
            if not self._package_unload(file_name):

                #-----------------------------------------------------------
                # In this case, specified package is already loaded, but
                # we failed to unload it. However, it should be totally fine
                # to delete corresponding file(s) of that package even if
                # it is loaded - at least in Python 2.
                # Therefore, we log this as a warning and continue deletion.

                log.warning("Package is loaded but failed to unload")

        # Find all files associated with this package name
        c_slash = "/"
        if package_path.endswith(c_slash):
            c_slash = ""
        path_file = os.path.abspath(package_path + c_slash + file_name)

        extensions = ["py", "pyc", "pyo"]
        file_names_found = []
        for file_ext_ind in extensions:
            if os.path.exists(path_file + "." + file_ext_ind):
                file_names_found.append(file_name + "." + file_ext_ind)

        if len(file_names_found) < 1:
            log.warning("No source or compiled file found for package: %s"
                        % file_name)
            return False

        # Create stash folder for delete packages
        stash_dir = package_path + c_slash + "stash"
        if not os.path.isdir(stash_dir):
            try:
                os.makedirs(stash_dir)
            except OSError:
                log.warning("Could not create stash directory: %s" % stash_dir)

        # Move files to stash directory
        flag_failed = False
        for file_name_found in file_names_found:
            file_name_src = package_path + c_slash + file_name_found
            try:
                os.rename(
                    file_name_src,
                    package_path + c_slash + "stash/" + file_name_found
                )
                log.debug("File moved to stash: %s" % file_name_found)
            except OSError:
                log.warning("Could not move file: %s" % file_name_src)
                try:
                    os.remove(file_name_src)
                    log.debug("File removed: %s" % file_name_found)
                except OSError:
                    log.error("Could not remove file: %s" % file_name_src)
                    flag_failed = True

        if flag_failed:
            log.warning("Some files failed to delete. See log for details")
        else:
            log.info("Package is deleted: %s" % file_name)
        return not flag_failed

    #-----------------------------------------------------------------------
    # This method is called to signal termination to all threads in package
    # manager, and calls metric handler to terminate its threads too.
    # Supposedly, it should be able to let Liota exit elegantly.

    def _terminate_all(self):
        global package_messenger_thread
        global package_messenger_pipe

        log.info("Shutting down package messenger...")
        if package_messenger_thread.isAlive():
            with open(package_messenger_pipe, "w") as fp:
                fp.write(
                    "terminate_messenger_but_you_should_not_do_this_yourself\n")

        log.info("Unloading packages...")
        if not self._package_unload_list(self._packages_loaded.keys()):
            log.error("Some packages failed to unload. See log for details")
            return False

        log.info("Terminating metric handler threads...")

        from liota.core.metric_handler import terminate as handler_terminate

        handler_terminate()
        return True


class PackageMessengerThread(Thread):
    """
    PackageMessengerThread does inter-process communication (IPC) to listen
    to commands casted by other processes (potentially from AirWatch, etc.)
    Current implementation of PackageMessengerThread blocks on a named pipe.
    """

    def __init__(self, pipe_file, name=None):
        Thread.__init__(self, name=name)
        self._pipe_file = pipe_file

        # Unblock previous writers
        BUFFER_SIZE = 65536
        ph = None
        try:
            ph = os.open(self._pipe_file, os.O_RDONLY | os.O_NONBLOCK)
            flags = fcntl.fcntl(ph, fcntl.F_GETFL)
            flags &= ~os.O_NONBLOCK
            fcntl.fcntl(ph, fcntl.F_SETFL, flags)
            while True:
                buffer = os.read(ph, BUFFER_SIZE)
                if not buffer:
                    break
        except OSError as err:
            import errno
            if err.errno == errno.EAGAIN or err.errno == errno.EWOULDBLOCK:
                pass  # It is supposed to raise one of these exceptions
            else:
                raise err
        finally:
            if ph:
                os.close(ph)

        self.flag_alive = True
        self.start()

    def run(self):
        global package_message_queue

        while self.flag_alive:
            with open(self._pipe_file, "r") as fp:
                for line in fp.readlines():
                    msg = line.split()
                    if len(msg) > 0:
                        if len(msg) == 1 and msg[0] == \
                                "terminate_messenger_but_you_should_not_do_this_yourself":
                            self.flag_alive = False
                            break
                        package_message_queue.put(msg)
        log.info("Thread exits: %s" % str(self.name))

#---------------------------------------------------------------------------
# Initialization should occur when this module is imported for first time.
# This method create queues and spawns PackageThread, which will loop on
# commands grabbed from a queue and load/unload packages as requested in
# those commands.


def initialize():
    global is_package_manager_initialized
    if is_package_manager_initialized:
        log.debug("Package manager is already initialized")
        return

    # Validate package path
    global package_path
    assert(isinstance(package_path, basestring))
    if os.path.isdir(package_path):
        try:
            os.listdir(package_path)
        except OSError:
            package_path = None
            log.error("Could not access package path")
            return
    else:
        log.debug("Could not find package path: " + package_path)
        try:
            os.makedirs(package_path)
            log.info("Created package path: " + package_path)
        except OSError:
            package_path = None
            log.error("Could not create package path")
            return

    # Validate package messenger pipe
    global package_messenger_pipe
    assert(isinstance(package_messenger_pipe, basestring))
    if os.path.exists(package_messenger_pipe):
        if stat.S_ISFIFO(os.stat(package_messenger_pipe).st_mode):
            pass
        else:
            log.error("Pipe path exists, but it is not a pipe")
            package_messenger_pipe = None
            return
    else:
        package_messenger_pipe_dir = os.path.dirname(package_messenger_pipe)
        if not os.path.isdir(package_messenger_pipe_dir):
            try:
                os.makedirs(package_messenger_pipe_dir)
                log.info("Created directory: " + package_messenger_pipe_dir)
            except OSError:
                package_messenger_pipe = None
                log.error("Could not create directory for messenger pipe")
                return
        try:
            os.mkfifo(package_messenger_pipe)
            log.info("Created pipe: " + package_messenger_pipe)
        except OSError:
            package_messenger_pipe = None
            log.error("Could not create messenger pipe")
            return
    assert(stat.S_ISFIFO(os.stat(package_messenger_pipe).st_mode))

    # Will not initialize package manager if package path is mis-configured
    if package_path is None:
        log.error("Package manager failed because package path is invalid")
        return

    # Initialization of package manager
    global package_message_queue
    if package_message_queue is None:
        package_message_queue = Queue()
    global package_lock
    if package_lock is None:
        package_lock = Lock()
    global package_thread
    if package_thread is None:
        package_thread = PackageThread(name="PackageThread")

    # PackageMessengerThread should start last because it triggers actions
    global package_messenger_thread
    if package_messenger_thread is None:
        if package_thread.isAlive():
            package_messenger_thread = \
                PackageMessengerThread(
                    pipe_file=package_messenger_pipe,
                    name="PackageMessengerThread"
                )
        else:
            log.warning("Package messenger will not start")

    # Mark package manager as initialized
    is_package_manager_initialized = True
    log.info("Package manager is initialized")

# Initialization of this module
if not is_package_manager_initialized:
    initialize()

log.debug("Package manager is imported")



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from abc import ABCMeta, abstractmethod
from liota.entities.entity import Entity

class Device(Entity):

    """
    Abstract base class for all devices (things).
    """
    __metaclass__ = ABCMeta

    #-----------------------------------------------------------------------
    # Constructor of Device is not made abstract, so developer can create a
    # plain device if that device does not have any specific method/data.
    #
    # It is possible to instantiate this class if no abstract method exists,
    # even if it has ABCMeta as its meta-class.
    #
    # Add @abstractmethod if this is not what we want.
    #
    def __init__(self, name, entity_id, entity_type="Device"):
        super(Device, self).__init__(
            name=name,
            entity_id=entity_id,
            entity_type=entity_type
        )



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#
import os
import sys
import json
import stat
import fcntl
import thread
import logging
from Queue import Queue
from threading import Thread

from liota.lib.utilities.utility import DiscUtilities
from liota.disc_listeners.discovery_listener import DiscoveryListener

log = logging.getLogger(__name__)

class NamedPipeListener(DiscoveryListener):
    """
    NamedPipeMessengerThread does inter-process communication (IPC), and
    blocks on a named pipe to listen to messages casted by devices.
    """

    def __init__(self, pipe_file, name=None, discovery=None):
        super(NamedPipeListener, self).__init__(name=name)
        if DiscUtilities().validate_named_pipe(pipe_file) == False:
            return None
        self._pipe_file = pipe_file
        self.msg_queue = Queue()
        self.discovery = discovery # backpoint to discovery obj

        # Unblock previous writers
        BUFFER_SIZE = 65536
        ph = None
        try:
            ph = os.open(self._pipe_file, os.O_RDONLY | os.O_NONBLOCK)
            flags = fcntl.fcntl(ph, fcntl.F_GETFL)
            flags &= ~os.O_NONBLOCK
            fcntl.fcntl(ph, fcntl.F_SETFL, flags)
            while True:
                buffer = os.read(ph, BUFFER_SIZE)
                if not buffer:
                    break
        except OSError as err:
            import errno
            if err.errno == errno.EAGAIN or err.errno == errno.EWOULDBLOCK:
                pass  # It is supposed to raise one of these exceptions
            else:
                raise err
        finally:
            if ph:
                os.close(ph)
        log.debug('NamedPipeListener is initialized')
        print 'NamedPipeListener is initialized'
        self.flag_alive = True
        self.start()

    def proc_dev_msg(self, payload):
        log.debug("named_pipe: proc_dev_msg")
        self.discovery.device_msg_process(payload)

    def run(self):
        log.info('NamedPipeListener is running')
        print 'NamedPipeListener is running'
        while self.flag_alive:
            with open(self._pipe_file, "r") as fp:
                data = ''
                for line in fp.readlines():
                    if line != 'END\n':
                        data += line
                    else:
                        try:
                            log.info('Received {0}'.format(data))
                            payload = json.loads(data)
                        except ValueError, err:
                            # json can't be parsed
                            log.error('Value: {0}, Error:{1}'.format(data, str(err)))
                            continue
                        Thread(target=self.proc_dev_msg, name="PipeMsgProc_Thread", args=(payload,)).start()
                        data = ''
        log.info("Thread exits: %s" % str(self.name))

    def clean_up(self):
        self.flag_alive = False


# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from abc import ABCMeta, abstractmethod
import logging

log = logging.getLogger(__name__)


class Filter:
    """
    Abstract base class for all Filters.

    Filtering can reduce network bandwidth by trimming off data that we are not interested in.  Also, most of the
    time systems will be working normally.  Sending all those normal data to DCC is not desired most of the time,
    as there is always storage and processing overhead involved.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def filter(self, v):
        """
        Child classes must implement appropriate filtering logic.

        :param v: Collected value by sampling function.
        :return: Filtered value or None
        """
        pass



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import Queue

import pint

from liota.dcc_comms.websocket_dcc_comms import WebSocketDccComms
from liota.dccs.dcc import RegistrationFailure
from liota.dccs.iotcc import IotControlCenter
from liota.device_comms.mqtt_device_comms import MqttDeviceComms
from liota.entities.devices.simulated_device import SimulatedDevice
from liota.entities.edge_systems.dk300_edge_system import Dk300EdgeSystem
from liota.entities.metrics.metric import Metric
from liota.lib.utilities.identity import Identity
from liota.lib.utilities.tls_conf import TLSConf

# getting values from conf file
config = {}
execfile('samplePropMqtt.conf', config)

# Create unit registry
ureg = pint.UnitRegistry()

# Store temperature values in Queue
kitchen_temperature_data = Queue.Queue()
living_room_temperature_data = Queue.Queue()


# Callback functions
# To put corresponding values in queue
def callback_kitchen_temp(client, userdata, message):
    kitchen_temperature_data.put(float(message.payload))


def callback_living_room_temp(client, userdata, message):
    living_room_temperature_data.put(float(message.payload))


# Extract data from Queue
def get_value(queue):
    return queue.get(block=True)


# ---------------------------------------------------------------------------------
# In this example, we demonstrate how data from two different Mqtt Channels (topics)
# can be collected and sent to IoTCC Dcc using Liota.
#
#
#                               IoTCC DCC
#                                  /|\
#                                   |
#                                   |
#                                   |  WebSocket
#                                   |
#                                   |
#                            Dell5kEdgeSystem
#                              /|\       /|\
#                               |         |
#             mqtt subscribe    |         |   mqtt subscribe
#          (temperature/kitchen)|         | (temperature/living-room)
#                               |         |
#                       --------------------------
#                      |                          |
#                      |        MQTT Broker       |
#                       --------------------------
#                       /|\                    /|\
#                        |                      |
#          mqtt publish  |                      |  mqtt publish
#  (temperature/kitchen) |                      | (temperature/living-room)
#                        |                      |
#                 Temperature Sensor       Temperature Sensor
#                   at Kitchen                at Living room
#
#
# Data streaming can be done from MQTT channel to IoTCC using LIOTA setting
# sampling_interval_sec to zero.
#
# Temperature values from sensor will be collected using MQTT channel and redirected
# to IoTCC data center component
# ------------------------------------------------------------------------------------


# MQTT connection setup to record kitchen and living room temperature values
def mqtt_subscribe():
    # Encapsulates Identity
    identity = Identity(config['broker_root_ca_cert'], config['broker_username'], config['broker_password'],
                              config['edge_system_cert_file'], config['edge_system_key_file'])
    # Encapsulate TLS parameters
    tls_conf = TLSConf(cert_required=config['cert_required'], tls_version=config['tls_version'],
                       cipher=config['cipher'])

    # Create MQTT connection object with required params
    mqtt_conn = MqttDeviceComms(url=config['BrokerIP'],  port=config['BrokerPort'], identity=identity,
                                tls_conf=tls_conf,
                                qos_details=None,
                                clean_session=True,
                                keep_alive=config['keep_alive'], enable_authentication=True)

    # Subscribe to channels : "temperature/kitchen" and "temperature/living-room" with preferred QoS level 0, 1 or 2
    # Provide callback function as a parameter for corresponding channel
    mqtt_conn.subscribe(config['MqttChannel1'], 1, callback_kitchen_temp)
    mqtt_conn.subscribe(config['MqttChannel2'], 1, callback_living_room_temp)


if __name__ == "__main__":

    #  Creating EdgeSystem
    edge_system = Dk300EdgeSystem(config['EdgeSystemName'])

    # Connect with MQTT broker using DeviceComms and subscribe to topics
    # Get kitchen and living room temperature values using MQTT channel
    mqtt_subscribe()

    # Create DCC object IoTCC using websocket transport
    # with UID and PASS
    iotcc = IotControlCenter(config['IotCCUID'], config['IotCCPassword'],
                             WebSocketDccComms(url=config['WebSocketUrl']))

    try:

        # Register Edge System with IoT control center
        reg_edge_system = iotcc.register(edge_system)

        # these call set properties on the Resource representing the IoT System
        # properties are a key:value store
        reg_edge_system.set_properties(config['SystemPropList'])

        # Create kitchen device object and register it on IoTCC
        # Add two device names in the configurations as DeviceName1 and DeviceName2
        kitchen_temperature_device = SimulatedDevice(name=config['DeviceName1'])
        reg_kitchen_temperature_device = iotcc.register(kitchen_temperature_device)

        iotcc.create_relationship(reg_edge_system, reg_kitchen_temperature_device)
        reg_kitchen_temperature_device.set_properties(config['DevicePropList'])

        # Metric Name
        metric_name_kitchen_temperature = "temperature.kitchen"

        # Create metric for kitchen temperature
        kitchen_temperature = Metric(
            name=metric_name_kitchen_temperature,
            unit=ureg.degC,
            interval=0,
            sampling_function=lambda: get_value(kitchen_temperature_data)
        )

        reg_kitchen_temp = iotcc.register(kitchen_temperature)
        iotcc.create_relationship(reg_kitchen_temperature_device, reg_kitchen_temp)
        reg_kitchen_temp.start_collecting()

        # Create living room device object and register it on IoTCC
        living_room_temperature_device = SimulatedDevice(name=config['DeviceName2'])
        reg_living_room_temperature_device = iotcc.register(living_room_temperature_device)

        iotcc.create_relationship(reg_edge_system, reg_living_room_temperature_device)
        reg_living_room_temperature_device.set_properties(config['DevicePropList'])

        # Metric Name
        metric_name_living_room_temperature = "temperature.living"

        # Create metric for living room temperature
        living_room_temperature = Metric(
            name=metric_name_living_room_temperature,
            unit=ureg.degC,
            interval=0,
            sampling_function=lambda:get_value(living_room_temperature_data)
        )

        reg_living_room_temp = iotcc.register(living_room_temperature)
        iotcc.create_relationship(reg_living_room_temperature_device, reg_living_room_temp)
        reg_living_room_temp.start_collecting()

    except RegistrationFailure:
        print "Registration to IOTCC failed"




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from pint import UnitRegistry
from temperusb import TemperHandler
from liota.dcc_comms.socket_comms import SocketDccComms
from liota.dccs.graphite import Graphite
from liota.entities.metrics.metric import Metric
from liota.entities.edge_systems.dk300_edge_system import Dk300EdgeSystem

# getting values from conf file
config = {}
execfile('sampleProp.conf', config)
ureg = UnitRegistry()
quantity = ureg.Quantity

# ---------------------------------------------------------------------------
# Below we showcase how easily SI units can be used in the code with help of
# Pint library which is part of Liota and how easily conversion of the units
# can be done such as Celsius to Fahrenheit or Kelvin as shown below with
# help of this library.

def getTemp():
    th = TemperHandler()
    devs = th.get_devices()
    t = devs[0].get_temperatures()
    temp = quantity(t[0]['temperature_c'], ureg.degC)
    return temp


def getTempDegC():
    temp_degC = getTemp()
    return temp_degC.magnitude


def getTempDegF():
    temp = getTemp()
    temp_degF = temp.to('degF')
    return temp_degF.magnitude


def getTempKelvin():
    temp = getTemp()
    temp_kelvin = temp.to('kelvin')
    return temp_kelvin.magnitude

# --------------------------------------------------------------------------
# In this example, we demonstrate how data from a simulated device generating
# random numbers can be directed to graphite data center component using Liota.
# The program illustrates the ease of use Liota brings to IoT application developers.

if __name__ == '__main__':

    edge_system = Dk300EdgeSystem(config['EdgeSystemName'])

    # Sending data to Graphite data center component
    # Socket is the underlying transport used to connect to the Graphite
    # instance
    graphite = Graphite(SocketDccComms(ip=config['GraphiteIP'],
                               port=config['GraphitePort']))
    graphite_reg_edge_system = graphite.register(edge_system)

    metric_name_temp_degC = config['MetricNameTempDegC']
    temp_metric_degC = Metric(
        name=metric_name_temp_degC,
        unit=ureg.degC,
        interval=33,
        aggregation_size=6,
        sampling_function=getTempDegC
    )
    reg_temp_metric_degC = graphite.register(temp_metric_degC)
    graphite.create_relationship(graphite_reg_edge_system, reg_temp_metric_degC)
    reg_temp_metric_degC.start_collecting()

    metric_name_temp_degF = config['MetricNameTempDegF']
    temp_metric_degF = Metric(
        name=metric_name_temp_degF,
        unit=ureg.degF,
        interval=62,
        aggregation_size=1,
        sampling_function=getTempDegF
    )
    reg_temp_metric_degF = graphite.register(temp_metric_degF)
    graphite.create_relationship(graphite_reg_edge_system, reg_temp_metric_degF)
    reg_temp_metric_degF.start_collecting()

    metric_name_temp_kelvin = config['MetricNameTempKelvin']
    temp_metric_kelvin = Metric(
        name=metric_name_temp_kelvin,
        unit=ureg.kelvin,
        interval=0,
        aggregation_size=1,
        sampling_function=getTempKelvin
    )
    reg_temp_metric_kelvin = graphite.register(temp_metric_kelvin)
    graphite.create_relationship(graphite_reg_edge_system, reg_temp_metric_kelvin)
    reg_temp_metric_kelvin.start_collecting()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from liota.core.package_manager import LiotaPackage

dependencies = ["edge_systems/dell5k/edge_system"]


class PackageClass(LiotaPackage):
    """
    This is a sample package which creates a IoTControlCenter DCC object and registers edge system on
    IoTCC over WebSocket to acquire "registered edge system", i.e. iotcc_edge_system.
    """

    def run(self, registry):
        import copy
        from liota.dccs.iotcc import IotControlCenter
        from liota.dcc_comms.websocket_dcc_comms import WebSocketDccComms
        from liota.dccs.dcc import RegistrationFailure

        # Acquire resources from registry
        # Creating a copy of edge_system object to keep original object "clean"
        edge_system = copy.copy(registry.get("edge_system"))

        # Get values from configuration file
        self.config_path = registry.get("package_conf")
        config = {}
        execfile(self.config_path + '/sampleProp.conf', config)

        # Initialize DCC object with transport
        self.iotcc = IotControlCenter(
            config['IotCCUID'], config['IotCCPassword'],
            WebSocketDccComms(url=config['WebSocketUrl'])
        )

        try:
            # Register edge system (gateway)
            self.iotcc_edge_system = self.iotcc.register(edge_system)
            """
            Use iotcc & iotcc_edge_system as common identifiers
            in the registry to easily refer the objects in other packages
            """
            registry.register("iotcc", self.iotcc)
            registry.register("iotcc_edge_system", self.iotcc_edge_system)
        except RegistrationFailure:
            print "EdgeSystem registration to IOTCC failed"
        self.iotcc.set_properties(self.iotcc_edge_system, config['SystemPropList'])

    def clean_up(self):
        # Get values from configuration file
        config = {}
        execfile(self.config_path + '/sampleProp.conf', config)

        #Unregister edge system
        if config['ShouldUnregisterOnUnload'] == "True":
            self.iotcc.unregister(self.iotcc_edge_system)
        self.iotcc.comms.client.close()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

from abc import abstractmethod
from numbers import Number
import logging

from liota.lib.utilities.utility import getUTCmillis
from liota.lib.utilities.filters.filter import Filter

log = logging.getLogger(__name__)


class WindowingScheme(Filter):
    """
    Applies windowing scheme to a Filter.

    It keeps track of a configurable time window.  Even if all values has been filtered out at the
    end of every time window, collected value is returned so that DCC is aware of it.
    """

    def __init__(self, filter_obj, window_size_sec):
        """
        :param filter_obj: Filter object.
        :param window_size_sec: Window size in seconds.
        """

        if not isinstance(filter_obj, Filter):
            raise TypeError("Filter Object is expected")
        if not isinstance(window_size_sec, Number) or window_size_sec < 0:
            log.error("window_size_sec must be a non negative number")
            raise ValueError("window_size_sec must be a non negative number")
        self.filter_obj = filter_obj
        self.window_size_sec = window_size_sec
        #  To track whether at-least one sample has been passed after filtering within a window
        self.sample_passed = False
        self.next_window_time = getUTCmillis() + (self.window_size_sec * 1000)

    def filter(self, v):
        """
        Applies filter and windowing scheme.

        :param v: Collected value by sampling function.
        :return: Filtered value or None
        """
        log.info("Applying windowing scheme")
        return self._window(v, self.filter_obj.filter(v))

    def _window(self, collected_value, filtered_value):
        """
        Windowing scheme.

        :param collected_value: Collected value by sampling function.
        :param filtered_value: Filtered value by sampling function.
        :return: Filtered value (or) collected value at the end of every time window.
        """
        # Next window time has elapsed.
        if getUTCmillis() >= self.next_window_time:
            #  At-least one sample has not passed so far during this window.
            if not self.sample_passed and filtered_value is None:
                self._set_next_window_time()
                log.info("Sending collected value for this window.")
                return collected_value

            # At-least one sample has (or will be) passed by now.
            else:
                self._set_next_window_time()
                return filtered_value  # Could be filtered-value or None

        # Next window time has not elapsed.
        else:
            if filtered_value is not None:
                self.sample_passed = True  # At-least one sample has passed during this window
            return filtered_value

    def _set_next_window_time(self):
        """
        Sets next time-window.
        :return: None
        """
        self.next_window_time += (self.window_size_sec * 1000)
        log.debug("Resetting window")
        self.sample_passed = False  # Resetting



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------#
#  Copyright © 2015-2016 VMware, Inc. All Rights Reserved.                    #
#                                                                             #
#  Licensed under the BSD 2-Clause License (the “License”); you may not use   #
#  this file except in compliance with the License.                           #
#                                                                             #
#  The BSD 2-Clause License                                                   #
#                                                                             #
#  Redistribution and use in source and binary forms, with or without         #
#  modification, are permitted provided that the following conditions are met:#
#                                                                             #
#  - Redistributions of source code must retain the above copyright notice,   #
#      this list of conditions and the following disclaimer.                  #
#                                                                             #
#  - Redistributions in binary form must reproduce the above copyright        #
#      notice, this list of conditions and the following disclaimer in the    #
#      documentation and/or other materials provided with the distribution.   #
#                                                                             #
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"#
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE  #
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE #
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE  #
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR        #
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF       #
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS   #
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN    #
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)    #
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF     #
#  THE POSSIBILITY OF SUCH DAMAGE.                                            #
# ----------------------------------------------------------------------------#

import logging
import os
import fcntl
import errno
import ConfigParser
from threading import Thread, Lock
from Queue import Queue
from time import sleep

from liota.lib.utilities.utility import LiotaConfigPath
from liota.lib.utilities.utility import DiscUtilities
from liota.dev_sims.named_pipe import NamedPipeSimulator
from liota.dev_sims.socket_clnt import SocketSimulator
from liota.dev_sims.mqtt import MqttSimulator

log = logging.getLogger(__name__)

is_discovery_simulator_initialized = False
# cmd related obj
cmd_message_queue = None
# simulator related obj
simulator_thread = None

if __name__ == "__main__":
    log.warning("Device Simulator is not supposed to run alone")

    import liota.core.discovery_simulator as actual_discovery_simulator

    log.debug("MainThread is waiting for interruption signal...")
    try:
        while not isinstance(
            actual_discovery_simulator.simulator_thread,
            actual_discovery_simulator.SimulatorThread
        ) or actual_discovery_simulator.simulator_thread.isAlive():
            sleep(1)
    except (KeyboardInterrupt, SystemExit):
        pass
    finally:
        # now only use named pipe for listening manage and statistic cmds
        if actual_discovery_simulator.cmd_message_queue is not None:
            actual_discovery_simulator.cmd_message_queue.put(["terminate"])
    log.info("Exiting Mainthread")
    exit()

class SimulatorThread(Thread):
    """
    SimulatorThread should be instantiated only once.
    Its instance serves core functionalities of Liota device simulator,
    including spawn multiple device simulator beaconing threads, etc.
    """

    def __init__(self, name=None):
        from liota.entities.edge_systems.dell5k_edge_system import Dell5KEdgeSystem

        Thread.__init__(self, name=name)
        # cmd related obj
        self.cmd_messenger_thread = None
        self.cmd_messenger_pipe = None
        # simulator related obj
        self.simulator_lock = None

        # simulator related configuration
        self.endpoint_list = {} # simulator list: (comm type, ip:port or folder)
        self.type_dcc_map = {} # Device Type to DCC mapping: (device type, dcc package name list (e.g., iotcc, iotcc_mqtt))
        self.type_key_map = {} # Device Type to Unique Key mapping: (device type, unique key)
        self.type_tuple_key_dcc_pkg = {} # Device Type to Tuple of (unique key, dcc_pkg)

        self._config = {} #key: cfg type, value: cfg list/info
        self._get_config_from_file() # extract configuration from liota.conf first
        self._executable_check() # check based on configuration, simulator thread is executable or not
        self._save_config()
        # create an edge system instance
        config_path = self._config['package_path']
        config = {}
        execfile(config_path + '/sampleProp.conf', config)
        self.edge_system_object = Dell5KEdgeSystem(config['EdgeSystemName'])

        # Initialization of simulator messenger queue and lock
        global cmd_message_queue
        cmd_message_queue = Queue()
        self.simulator_lock = Lock()

        self._simulators = {}  # key: beacon type, value: thread ref
        self.flag_alive = True
        self.start()

    def _get_config_from_file(self):
        # Parse Liota configuration file
        config = ConfigParser.RawConfigParser()
        config.optionxform=str
        fullPath = LiotaConfigPath().get_liota_fullpath()
        if fullPath != '':
            try:
                if config.read(fullPath) != []:
                    try:
                        # retrieve device info file storage directory
                        self.dev_file_path = config.get('IOTCC_PATH', 'dev_file_path')
                    except ConfigParser.ParsingError as err:
                        log.error('Could not parse log config file' + err)
                        exit(-4)
                    if not os.path.exists(self.dev_file_path):
                        try:
                            os.makedirs(self.dev_file_path)
                        except OSError as exc:  # Python >2.5
                            if exc.errno == errno.EEXIST and os.path.isdir(self.dev_file_path):
                                pass
                            else:
                                log.error('Could not create device file storage directory')
                                exit(-4)
                    try:
                        # retrieve simulator cmd pip
                        self.cmd_messenger_pipe = os.path.abspath(
                            config.get('DEVSIM_CFG', 'devsim_cmd_msg_pipe')
                        )
                        # retrieve package path (later use config file to register device)
                        self.package_path = os.path.abspath(
                            config.get('PKG_CFG', 'pkg_path')
                        )
                        # retrieve endpoint list
                        tmp_list = config.items('DISC_ENDPOINT_LIST')
                        for key, value in tmp_list[:]:
                            if value is None or value == "None":
                                continue
                            if key.find('disc_msg_pipe') != -1:
                                self.simulator_messenger_pipe = os.path.abspath(value)
                                if self.simulator_messenger_pipe is not None:
                                    self.endpoint_list[key] = self.simulator_messenger_pipe
                            elif key.find('socket') != -1:
                                self.endpoint_list[key] = value
                            elif key.find('mqtt') != -1:
                                # retrieve mqtt configurations
                                mqtt_cfg_dict = dict(config.items('DISC_MQTT_CFG'))
                                # make broker_ip:port:topic included inside mqtt_cfg_dict
                                mqtt_cfg_dict["broker_ip_port_topic"] = value
                                self.endpoint_list[key] = mqtt_cfg_dict
                            elif key.find('coap') != -1:
                                self.endpoint_list[key] = value
                            else:
                                log.error(key + ' is currently not supported!')
                        for key in self.endpoint_list.iterkeys():
                            log.debug("endpoint_list:(%s : %s)\n" % (key, self.endpoint_list[key]))

                        # retrieve device type to unique key mapping list
                        tmp_list = config.items('DEVICE_TYPE_TO_UNIQUEKEY_MAPPING')
                        for key, value in tmp_list[:]:
                            if value is None or value == "None":
                                continue
                            self.type_key_map[key] = value
                        for key in self.type_key_map.iterkeys():
                            log.debug("type_key_map:(%s : %s)\n" % (key, self.type_key_map[key]))

                        # retrieve device type to DCC mapping list
                        tmp_list = config.items('DEVICE_TYPE_TO_DCC_MAPPING')
                        for key, value in tmp_list[:]:
                            if value is None or value == "None":
                                continue
                            tmp_list2 = []
                            tmp_list2 = [x.strip() for x in value.split(',')]
                            self.type_dcc_map[key] = tmp_list2
                            self.type_tuple_key_dcc_pkg[key] = (self.type_key_map[key], tmp_list2)
                        for key in self.type_dcc_map.iterkeys():
                            log.debug("type_dcc_map:(%s : %s)\n" % (key, self.type_dcc_map[key]))
                        for key in self.type_tuple_key_dcc_pkg.iterkeys():
                            log.debug("type_tuple_key_dcc_pkg:(%s : %s)\n" % (key, self.type_tuple_key_dcc_pkg[key]))
                    except ConfigParser.ParsingError:
                        log.error('Could not parse log config file')
                        exit(-4)
                else:
                    raise IOError('Could not open configuration file: ' + fullPath)
            except IOError:
                raise IOError('Could not open configuration file: ' + fullPath)
        else:
            # missing config file
            log.error('liota.conf file missing')

        assert(isinstance(self.cmd_messenger_pipe, basestring))

    def _executable_check(self):
        if DiscUtilities().validate_named_pipe(self.cmd_messenger_pipe) == False:
            return None
        # Will not initialize device simulator if simulator list is empty
        if len(self.endpoint_list) is 0:
            log.error("Device simulator failed because simulator list is empty")
            return

    def _save_config(self):
        # cmd messenger related
        self._config['cmd_msg_pipe'] = self.cmd_messenger_pipe
        # simulator related
        # simulator list: (comm type, ip:port or folder)
        self._config['endpoint_list'] = self.endpoint_list
        # Device Type to DCC mapping: (device type, dcc package name list (e.g., iotcc, iotcc_mqtt))
        self._config['type_dcc_map'] = self.type_dcc_map
        # Device Type to Unique Key mapping: (device type, unique key)
        self._config['type_key_map'] = self.type_key_map
        # Device Type to Tuple of (unique key, dcc, dcc_pkg)
        self._config['type_tuple_key_dcc_pkg'] = self.type_tuple_key_dcc_pkg
        # package path
        self._config['package_path'] = self.package_path
        # device info storage path
        self._config['dev_file_path']= self.dev_file_path

    #-----------------------------------------------------------------------
    # This method is used to handle listing commands

    def _cmd_handler_list(self, parameter):
        # configurations
        if parameter == "configurations" or parameter == "cfg":
            stats = ["n/a", "n/a", "n/a", "n/a", "n/a", "n/a", "n/a"]
            stats[0] = str(self._config['cmd_msg_pipe']) + '\n\t'
            tmp = ''
            endpoint_list = self._config['endpoint_list']
            for key in endpoint_list.iterkeys():
                tmp += str(key) + ': ' + str(endpoint_list[key]) + '\n\t\t'
            stats[1] = tmp

            tmp = ''
            type_dcc_map = self._config['type_dcc_map']
            for key in type_dcc_map.iterkeys():
                tmp += str(key) + ': ' + str(type_dcc_map[key]) + '\n\t\t'
            stats[2] = tmp

            tmp = ''
            type_key_map = self._config['type_key_map']
            for key in type_key_map.iterkeys():
                tmp += str(key) + ': ' + str(type_key_map[key]) + '\n\t\t'
            stats[3] = tmp

            tmp = ''
            type_tuple_key_dcc_pkg = self._config['type_tuple_key_dcc_pkg']
            for key in type_tuple_key_dcc_pkg.iterkeys():
                tmp += str(key) + ': ' + str(type_tuple_key_dcc_pkg[key]) + '\n\t\t'
            stats[4] = tmp

            stats[5] = str(self._config['package_path']) + '\n\t'
            stats[6] = str(self._config['dev_file_path']) + '\n\t'
            log.warning(("List of configurations - \n\t"
                        + "cmd_msg_pipe: %s\n\t"
                        + "endpoint_list: %s\n\t"
                        + "type_dcc_map: %s\n\t"
                        + "type_key_map: %s\n\t"
                        + "type_tuple_key_dcc_pkg: %s\n\t"
                        + "dev_file_path: %s\n\t"
                        + "package_path: %s"
                        ) % tuple(stats))
            return
        # simulator threads
        if parameter == "threads" or parameter == "th":
            import threading

            log.warning("Active threads - \n\t%s"
                        % "\n\t".join(map(
                            lambda tref: "%s: %016x %s %s" % (
                                tref.name,
                                tref.ident,
                                type(tref).__name__.split(".")[-1],
                                tref.isAlive()
                            ),
                            sorted(
                                threading.enumerate(),
                                key=lambda tref: tref.ident
                            )
                        ))
                        )
            return
        log.warning("Unsupported list")

    #-----------------------------------------------------------------------
    # This method is used to handle statistical commands

    def _cmd_handler_stat(self, parameter):
        if parameter == "threads" or parameter == "th":
            import threading

            log.warning("Count of active threads: %d"
                        % threading.active_count())
            return
        log.warning("Unsupported stat")

    #-----------------------------------------------------------------------
    # This method will loop on message queue and select methods to call with
    # respect to commands received.

    def run(self):
        from liota.dev_sims.coap import CoapSimulator

        endpoint_list = self._config['endpoint_list']

        # spin listening threads according to endpoint_list extracted config
        for key in endpoint_list.iterkeys():
            value = endpoint_list[key]
            log.debug("Endpoint:{0}:{1}".format(key, value))
            if value is None or value == "None":
                continue
            if key.find('disc_msg_pipe') != -1:
                pipe_thread = NamedPipeSimulator(pipe_file=value,
                        name=key+"_Thread", simulator=self)
                if pipe_thread is not None:
                    self._simulators[key] = pipe_thread
            if key.find('socket') != -1:
                socket_thread = SocketSimulator(ip_port=value,
                        name=key+"_Thread", simulator=self)
                if socket_thread is not None:
                    self._simulators[key] = socket_thread
            if key.find('mqtt') != -1:
                mqtt_thread = MqttSimulator(mqtt_cfg=value,
                        name=key+"_Thread", simulator=self)
                if mqtt_thread is not None:
                    self._simulators[key] = mqtt_thread
            if key.find('coap') != -1:
                coap_thread = CoapSimulator(ip_port=value,
                        name=key+"_Thread", simulator=self)
                if coap_thread is not None:
                    self._simulators[key] = coap_thread

        # Listen on message queue for management or statistic commands
        global cmd_message_queue
        self.cmd_messenger_thread = \
            CmdMessengerThread(
                pipe_file=self.cmd_messenger_pipe,
                name="CmdMessengerThread",
                cmd_queue = cmd_message_queue
            )
        # currently simulators are only allocated once based on info in liota.conf
        while self.flag_alive:
            msg = cmd_message_queue.get()
            log.info("Got message in cmd messenger queue: %s"
                     % " ".join(msg))
            if not isinstance(msg, tuple) and not isinstance(msg, list):
                raise TypeError(type(msg))

            # Switch on message content (command), determine what to do
            command = msg[0]
            if command == "list":
                with self.simulator_lock:
                    if len(msg) != 2:
                        log.warning("Invalid format of command: %s" % command)
                        continue
                    self._cmd_handler_list(msg[1])
            elif command == "stat":
                with self.simulator_lock:
                    if len(msg) != 2:
                        log.warning("Invalid format of command: %s" % command)
                        continue
                    self._cmd_handler_stat(msg[1])
            elif command == "terminate":
                if self._terminate_all():
                    self.flag_alive = False
                    break
            else:
                log.warning("Unsupported command is dropped")

        log.info("Thread exits: %s" % str(self.name))

    #-----------------------------------------------------------------------
    # This method is called to signal termination to all threads in device
    # simulator, and calls message processor to terminate its threads too.
    # Supposedly, it should be able to let Liota exit elegantly.

    def _terminate_all(self):
        log.info("Shutting down device simulator thread...")
        if (self.cmd_messenger_thread is not None) and (self.cmd_messenger_thread.isAlive()):
            with open(self.cmd_messenger_pipe, "w") as fp:
                fp.write(
                    "terminate_messenger_but_you_should_not_do_this_yourself\n")

        log.info("terminate simulator threads...")
        for k, v in self._simulators.items():
            if v is not None:
                v.clean_up()
                del self._simulators[k]
        return True


class CmdMessengerThread(Thread):
    """
    CmdMessengerThread does inter-process communication (IPC) to listen
    to commands casted by other processes.
    Current implementation of CmdMessengerThread blocks on a named pipe.
    """

    def __init__(self, pipe_file, name=None, cmd_queue=None):
        Thread.__init__(self, name=name)
        self._pipe_file = pipe_file
        if cmd_queue is None:
            log.error("Thread exits: %s (because cmd_queue is None)" % str(self.name))
            return
        global cmd_message_queue
        cmd_message_queue = cmd_queue

        # Unblock previous writers
        BUFFER_SIZE = 65536
        ph = None
        try:
            ph = os.open(self._pipe_file, os.O_RDONLY | os.O_NONBLOCK)
            flags = fcntl.fcntl(ph, fcntl.F_GETFL)
            flags &= ~os.O_NONBLOCK
            fcntl.fcntl(ph, fcntl.F_SETFL, flags)
            while True:
                buf = os.read(ph, BUFFER_SIZE)
                if not buf:
                    break
        except OSError as err:
            if err.errno == errno.EAGAIN or err.errno == errno.EWOULDBLOCK:
                pass  # It is supposed to raise one of these exceptions
            else:
                raise err
        finally:
            if ph:
                os.close(ph)

        self.flag_alive = True
        self.start()

    def run(self):

        while self.flag_alive:
            with open(self._pipe_file, "r") as fp:
                for line in fp.readlines():
                    msg = line.split()
                    if len(msg) > 0:
                        if len(msg) == 1 and msg[0] == \
                                "terminate_messenger_but_you_should_not_do_this_yourself":
                            self.flag_alive = False
                            break
                        global cmd_message_queue
                        cmd_message_queue.put(msg)
        log.info("Thread exits: %s" % str(self.name))

#---------------------------------------------------------------------------
# Initialization should occur when this module is imported for first time.
# This method create queues and spawns SimulatorThread, which will spins up
# beaconing threads for listed endpoint in liota.conf for simulated devices.

def initialize():
    global is_discovery_simulator_initialized
    if is_discovery_simulator_initialized:
        log.debug("Discovery simulator is already initialized")
        return

    # Initialization of device simulator
    global simulator_thread
    simulator_thread = SimulatorThread(name="SimulatorThread")

    # Mark discovery simulator as initialized
    is_discovery_simulator_initialized = True
    log.info("Device simulator is initialized")

# Initialization of this module
if not is_discovery_simulator_initialized:
    initialize()

log.debug("Discovery simulator is imported")
